﻿/**
 * Sunshine's simple big unsigned integer library
 * ----------------------------------------------
 * Version 1.0 / 2020-08-19
 *
 * Simple console application to test my own big (unsigned) integer library implementation.
 *
 * www.sunshine2k.de || www.bastian-molkenthin.de
 */

using SunBigUInteger.SunBigUInt;
using SunBigUInteger.SunBigUIntConcrete;
using SunBigUInteger.SunBigUIntTests;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Text;
using System.Threading.Tasks;

namespace SunBigUInteger
{
    class Program
    {
        static void Main(string[] args)
        {
            PrintHeader();

#if TRUE
            Console.WriteLine();
            RunSunBigUInt10GenericTestcases();

            Console.WriteLine("");
            RunSunBigUInt10SpecificTestcases();

            Console.WriteLine();
            RunSunBigUInt256GenericTestcases();

            Console.WriteLine();
            RunSunBigUInt256SpecificTestcases();

            Console.WriteLine("Press key to quit");
            Console.ReadLine();

#else
            /* Manual playground for debugging */
            SunBigUInt10Concrete c1 = SunBigUInt10Concrete.FromLong(300) + SunBigUInt10Concrete.FromString("912");
            Console.WriteLine(c1);

            SunBigUInt10Concrete c2 = c1 * 02;
            Console.WriteLine(c2);

            SunBigUInt10Concrete c3 = 24 / 4;
            Console.WriteLine(c3);

            SunBigUInt10Concrete r;
            SunBigUInt10Concrete c4 = SunBigUInt10Concrete.DivRem(24, 4, out r);

            Console.WriteLine("Press key to quit");
            Console.ReadLine();
#endif
        }

        private static void PrintHeader()
        {
            Console.Out.WriteLine("SunBigUInteger -  Arithmetic operations for unsigned integers of arbitrary precision.");
            Console.Out.WriteLine("Version: 1.0");
            Console.Out.WriteLine("www.sunshine2k.de | www.bastian - molkenthin.de");
            Console.Out.WriteLine("---------------------");
        }

        private static void PrintTestVerdict(int numOfFailedTestcases)
        {
            Console.WriteLine();
            if (numOfFailedTestcases == 0)
            {
                Console.WriteLine("--> Test verdict: OK");
            }
            else
            {
                Console.WriteLine("--> Test verdict: FAILED - " + numOfFailedTestcases + " test cases :-(");
            }
        }

        private static void RunSunBigUInt10GenericTestcases()
        {
            TestHandler<SunBigUInt10> app = new TestHandler<SunBigUInt10>();
            Console.WriteLine("Running tests: SunBigUInt10 (Generic)");
            SunBigUIntTestSuite <SunBigUInt10> sunBigUInt10Tests = new SunBigUIntTestSuite<SunBigUInt10>(app);
            sunBigUInt10Tests.Run<SunBigUInt10>(typeof(SunBigUInt10));
            PrintTestVerdict(sunBigUInt10Tests.NumOfFailedTestcases);
        }

        private static void RunSunBigUInt10SpecificTestcases()
        {
            TestHandler app = new TestHandler();
            Console.WriteLine("Running tests: SunBigUInt10 (Specific)");
            SunBigUInt10SpecificTests sunBigUInt10Tests = new SunBigUInt10SpecificTests(app);
            sunBigUInt10Tests.Run();
            PrintTestVerdict(sunBigUInt10Tests.NumOfFailedTestcases);
        }

        private static void RunSunBigUInt256GenericTestcases()
        {
            TestHandler<SunBigUInt256> app = new TestHandler<SunBigUInt256>();
            Console.WriteLine("Running tests: SunBigUInt256 (Generic)");
            SunBigUIntTestSuite<SunBigUInt256> sunBigUInt256Tests = new SunBigUIntTestSuite<SunBigUInt256>(app);
            sunBigUInt256Tests.Run<SunBigUInt256>(typeof(SunBigUInt256));
            PrintTestVerdict(sunBigUInt256Tests.NumOfFailedTestcases);
        }

        private static void RunSunBigUInt256SpecificTestcases()
        {
            TestHandler app = new TestHandler();
            Console.WriteLine("Running tests: SunBigUInt256 (Specific)");
            SunBigUInt256SpecificTests sunBigUInt256Tests = new SunBigUInt256SpecificTests(app);
            sunBigUInt256Tests.Run();
            PrintTestVerdict(sunBigUInt256Tests.NumOfFailedTestcases);
        }

    }

    class TestHandler<TClass> : ITestHandler<TClass> where TClass : SunBigUIntAbstract<TClass>
    {
        public void TestGroupStart(string name)
        {
            Console.Write(name + ": ");
        }

        public void TestGroupEnd()
        {
            Console.WriteLine();
        }

        public void TestStepFail(string message)
        {
            Console.WriteLine();
            Console.WriteLine("F : " + message);
        }

        public void TestStepSuccess()
        {
            Console.Write('x');
        }

        public void TestStepFail(BigInteger expectedInt, SunBigUInt10 actualInt)
        {
            Console.WriteLine();
            Console.WriteLine("F: Expected: " + expectedInt + " Actual: " + actualInt);
        }

        public void TestStepFail(BigInteger expectedInt, SunBigUInt256 actualInt)
        {
            Console.WriteLine();
            Console.WriteLine("F: Expected: " + expectedInt + " Actual: " + actualInt);
        }

        public void TestStepFail<T1>(BigInteger expectedInt, T1 actualInt)
        {
            Console.WriteLine();
            Console.WriteLine("F: Expected: " + expectedInt + " Actual: " + actualInt);
        }
    }

    class TestHandler : ITestHandler
    {
        public void TestGroupStart(string name)
        {
            Console.Write(name + ": ");
        }

        public void TestGroupEnd()
        {
            Console.WriteLine();
        }

        public void TestStepFail(string message)
        {
            Console.WriteLine();
            Console.WriteLine("F : " + message);
        }

        public void TestStepSuccess()
        {
            Console.Write('x');
        }

        public void TestStepFail(BigInteger expectedInt, SunBigUInt10 actualInt)
        {
            Console.WriteLine();
            Console.WriteLine("F: Expected: " + expectedInt + " Actual: " + actualInt);
        }

        public void TestStepFail(BigInteger expectedInt, SunBigUInt256 actualInt)
        {
            Console.WriteLine();
            Console.WriteLine("F: Expected: " + expectedInt + " Actual: " + actualInt);
        }
    }
}
