﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SunBigUInteger.SunBigUInt
{
    class SunBigUInt10 : SunBigUIntAbstract<SunBigUInt10>
    {
        private byte[] digits;

        public SunBigUInt10() { }

        private SunBigUInt10(byte[] digitsTemp)
        {
            int firstPosNonZero = digitsTemp.Length - 1;
            bool isZero = true;
            for (int i = digitsTemp.Length - 1; i >= 0; i--)
            {
                if (digitsTemp[i] != 0)
                {
                    firstPosNonZero = i;
                    isZero = false;
                    break;
                }
            }

            if (isZero && (digitsTemp.Length > 1))
            {
                // all digits are zero, but have more than one element, just return valid zero element
                this.digits = new byte[1];
                this.digits[0] = 0;
                return;
            }

            this.digits = new byte[firstPosNonZero + 1];
            for (int i = 0; i <= firstPosNonZero; i++)
            {
                this.digits[i] = digitsTemp[i];
            }
        }

        public override SunBigUInt10 FromLong(ulong number)
        {
            return SunBigUInt10.SFromLong(number);
        }

        public static SunBigUInt10 SFromLong(ulong number)
        {
            if (number < 0)
            {
                throw new ArgumentOutOfRangeException("Negative numbers not supported.");
            }

            SunBigUInt10 n = new SunBigUInt10();
            if (number == 0)
            {
                n.digits = new byte[1];
                return n;
            }

            // count number of digits
            ulong numberCopy = number;
            int numDigits = 0;
            while (numberCopy > 0)
            {
                numberCopy /= 10;
                numDigits++;
            }

            n.digits = new byte[numDigits];
            int i = 0;
            while (number > 0)
            {
                n.digits[i] = (byte)(number % 10);
                number /= 10;
                i++;
            }

            return n;
        }

        public static SunBigUInt10 SFromString(string number)
        {
            if (String.IsNullOrEmpty(number))
            {
                throw new ArgumentOutOfRangeException("String is NULL or empty.");
            }
            // remove leading zeros but avoid empty string
            number = number.TrimStart('0');
            number = number.TrimStart(' ');
            number = number.TrimEnd(' ');
            number = number.Length > 0 ? number : "0";

            SunBigUInt10 n = new SunBigUInt10
            {
                digits = new byte[number.Length]
            };

            for (int i = 0; i < number.Length; i++)
            {
                if (!Char.IsDigit(number[number.Length - i - 1]))
                {
                    throw new ArgumentException("Invalid decimal digit: " + number[number.Length - i - 1]);
                }
                n.digits[i] = (byte)(number[number.Length - i - 1] - 0x30);
            }
            return n;
        }

        public override SunBigUInt10 FromString(string number)
        {
            return SunBigUInt10.SFromString(number);
        }

        public override SunBigUInt10 FromHexString(string number)
        {
            if (String.IsNullOrEmpty(number))
            {
                throw new ArgumentOutOfRangeException("String is NULL or empty.");
            }

            if (!number.StartsWith("0x"))
            {
                throw new ArgumentException("Hexadecimal string does not start with '0x'");
            }

            // remove leading '0x'
            number = number.Substring(2);
            // remove leading zeros but avoid empty string
            number = number.TrimStart('0');
            number = number.TrimStart(' ');
            number = number.TrimEnd(' ');
            number = number.Length > 0 ? number : "0";

            SunBigUInt10 n = 0;
            for (int i = 0; i < number.Length; i++)
            {
                char c = number[i];
                int digitVal;

                if ((c >= '0') && (c <= '9'))
                {
                    digitVal = c - '0';
                }
                else if ((c >= 'a') && (c <= 'f'))
                {
                    digitVal = c - 'a' + 10;
                }
                else if ((c >= 'A') && (c <= 'F'))
                {
                    digitVal = c - 'A' + 10;
                }
                else
                {
                    throw new ArgumentException("Invalid hexadecimal digit: " + c);
                }

                n = (n * (SunBigUInt10)16) + (SunBigUInt10)digitVal;
            }

            return n;
        }

        public static implicit operator SunBigUInt10(int number)
        {
            return SunBigUInt10.SFromLong((ulong)number);
        }

        public static implicit operator SunBigUInt10(ulong number)
        {
            return SunBigUInt10.SFromLong(number);
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            for (int i = this.digits.Length - 1; i >= 0; i--)
            {
                sb.Append(digits[i].ToString());
            }
            return sb.ToString();
        }

        /* Addition */
        public override SunBigUInt10 Add(SunBigUIntAbstract<SunBigUInt10> c1)
        {
            return SunBigUInt10.Add((SunBigUInt10)this, (SunBigUInt10)c1);
        }

        public static SunBigUInt10 Add(SunBigUInt10 a, SunBigUInt10 b)
        {
            byte[] temp = new byte[Math.Max(a.digits.Length, b.digits.Length) + 1];
            byte carry = 0;
            for (int i = 0; i < temp.Length; i++)
            {
                int anum = i < a.digits.Length ? a.digits[i] : 0;
                int bnum = i < b.digits.Length ? b.digits[i] : 0;
                byte sumNum = (byte)(anum + bnum + carry);
                if (sumNum >= 10)
                {
                    sumNum -= 10;
                    carry = 1;
                }
                else
                {
                    carry = 0;
                }
                temp[i] = sumNum;
            }
            return new SunBigUInt10(temp);
        }

        /* Subtraction */
        public override SunBigUInt10 Sub(SunBigUIntAbstract<SunBigUInt10> c1)
        {
            return SunBigUInt10.Sub((SunBigUInt10)this, (SunBigUInt10)c1);
        }

        public static SunBigUInt10 Sub(SunBigUInt10 a, SunBigUInt10 b)
        {
            if (b > a)
            {
                throw new OverflowException("b is greater than a.");
            }

            byte[] temp = new byte[a.digits.Length];
            byte borrow = 0;
            for (int i = 0; i < temp.Length; i++)
            {
                int anum = a.digits[i];
                int bnum = i < b.digits.Length ? b.digits[i] : 0;
                byte diffNum;
                if (anum < (bnum + borrow))
                {
                    // overflow
                    diffNum = (byte)(anum + 10 - bnum - borrow);
                    borrow = 1;
                }
                else
                {
                    diffNum = (byte)(anum - bnum - borrow);
                    borrow = 0;
                }
                temp[i] = diffNum;
            }

            return new SunBigUInt10(temp);
        }


        /* Multiplication */
        public override SunBigUInt10 Mul(SunBigUIntAbstract<SunBigUInt10> c1)
        {
            return SunBigUInt10.Mul((SunBigUInt10)this, (SunBigUInt10)c1);
        }

        public static SunBigUInt10 Mul(SunBigUInt10 a, SunBigUInt10 b)
        {
            byte[] tmpArray = new byte[a.digits.Length + b.digits.Length];

            int i, j;
            for (j = 0; j < b.digits.Length; j++)
            {
                byte carry = 0;
                for (i = 0; i < a.digits.Length; i++)
                {
                    // note: max carry is 9*9 + 9 = 90 -> fits in a byte
                    carry += (byte)(a.digits[i] * b.digits[j] + tmpArray[i + j]);
                    tmpArray[i + j] = (byte)(carry % 10);
                    carry /= 10;
                }

                if (carry != 0)
                {
                    tmpArray[i + j] = (byte)(carry % 10);
                }
            }

            return new SunBigUInt10(tmpArray);
        }

        /* Division */
        public override SunBigUInt10 Div(SunBigUIntAbstract<SunBigUInt10> c1)
        {
            return SunBigUInt10.Div((SunBigUInt10)this, (SunBigUInt10)c1);
        }

        public static SunBigUInt10 Div(SunBigUInt10 a, SunBigUInt10 b)
        {
            SunBigUInt10 q, r;
            if (b == (SunBigUInt10)0)
            {
                throw new DivideByZeroException("b is 0");
            }

            DivRem(a, b, out q, out r);
            return q;
        }

        public override void DivRem(SunBigUIntAbstract<SunBigUInt10> a, SunBigUIntAbstract<SunBigUInt10> b, 
                                    out SunBigUIntAbstract<SunBigUInt10> q, out SunBigUIntAbstract<SunBigUInt10> r)
        {
            SunBigUInt10 q2;
            SunBigUInt10 r2;
            SunBigUInt10.DivRem((SunBigUInt10)a, (SunBigUInt10)b, out q2, out r2);
            q = q2;
            r = r2;
        }

        public override SunBigUInt10 DivRem(SunBigUIntAbstract<SunBigUInt10> a, SunBigUIntAbstract<SunBigUInt10> b, out SunBigUIntAbstract<SunBigUInt10> r)
        {
            if (b == (SunBigUInt10)0)
            {
                throw new DivideByZeroException("b is 0");
            }

            SunBigUIntAbstract<SunBigUInt10> q;
            DivRem(a, b, out q, out r);
            return (SunBigUInt10)q;
        }


        public static SunBigUInt10 DivRem(SunBigUInt10 a, SunBigUInt10 b, out SunBigUInt10 r)
        {
            if (b == (SunBigUInt10)0)
            {
                throw new DivideByZeroException("b is 0");
            }

            SunBigUInt10 q;
            DivRem(a, b, out q, out r);
            return q;
        }

        public void DivRem(SunBigUInt10 b, out SunBigUInt10 q, out SunBigUInt10 r)
        {
            SunBigUInt10.DivRem(this, b, out q, out r);
        }

        private static void DivRem(SunBigUInt10 a, SunBigUInt10 b, out SunBigUInt10 q, out SunBigUInt10 r)
        {
            if (a < b)
            {
                q = 0; r = a;
            }
            else
            {
                SunBigUInt10 qTmp, rTmp;
                DivRem(a, (SunBigUInt10)2 * b, out qTmp, out rTmp);
                if (rTmp < b)
                {
                    q = (SunBigUInt10)2 * qTmp;
                    r = rTmp;
                }
                else
                {
                    q = (SunBigUInt10)2 * qTmp + (SunBigUInt10)1;
                    r = rTmp - b;
                }
            }
        }

        /* Modulus */
        public override SunBigUInt10 Mod(SunBigUIntAbstract<SunBigUInt10> c1)
        {
            return SunBigUInt10.Mod((SunBigUInt10)this, (SunBigUInt10)c1);
        }
        public static SunBigUInt10 Mod(SunBigUInt10 a, SunBigUInt10 b)
        {
            SunBigUInt10 q, r;
            if (b == (SunBigUInt10)0)
            {
                throw new DivideByZeroException("b is 0");
            }

            DivRem(a, b, out q, out r);
            return r;
        }

        /* Pow / ModPow */
        public override SunBigUInt10 Pow(SunBigUIntAbstract<SunBigUInt10> a, SunBigUIntAbstract<SunBigUInt10> b)
        {
            return SunBigUInt10.Pow((SunBigUInt10)a, (SunBigUInt10)b);
        }
        public override SunBigUInt10 Pow(SunBigUIntAbstract<SunBigUInt10> a, int b)
        {
            return SunBigUInt10.Pow((SunBigUInt10)a, (SunBigUInt10)b);
        }
        public static SunBigUInt10 Pow(SunBigUInt10 a, SunBigUInt10 b)
        {
            SunBigUInt10 res = 1;
            while (b > (SunBigUInt10)0)
            {
                if (b.IsOdd())
                {
                    res *= a;
                }
                b /= (SunBigUInt10)2;
                a = a * a;
            }

            return res;
        }

        public static SunBigUInt10 PowSimple(SunBigUInt10 a, SunBigUInt10 b)
        {
            SunBigUInt10 res = 1;
            for (SunBigUInt10 i = 0; i < b; i++)
            {
                res = res * a;
            }
            return res;
        }

        public override SunBigUInt10 ModPow(SunBigUIntAbstract<SunBigUInt10> value, SunBigUIntAbstract<SunBigUInt10> exponent, SunBigUIntAbstract<SunBigUInt10> modulus)
        {
            return SunBigUInt10.ModPow((SunBigUInt10)value, (SunBigUInt10)exponent, (SunBigUInt10)modulus);
        }
        public override SunBigUInt10 ModPow(int value, int exponent, int modulus)
        {
            return SunBigUInt10.ModPow((SunBigUInt10)value, (SunBigUInt10)exponent, (SunBigUInt10)modulus);
        }
        public static SunBigUInt10 ModPow(SunBigUInt10 value, SunBigUInt10 exponent, SunBigUInt10 modulus)
        {
            if (modulus == (SunBigUInt10)1)
            {
                return 0;
            }

            SunBigUInt10 c = 1;
            for (SunBigUInt10 i = 0; i < exponent; i++)
            {
                c = (c * value).Mod(modulus);
            }
            return c;
        }

        /* Increment / Decrement */
        public override SunBigUInt10 Inc(SunBigUIntAbstract<SunBigUInt10> a)
        {
            return (SunBigUInt10)a + (SunBigUInt10)1;
        }
        public override SunBigUInt10 Dec(SunBigUIntAbstract<SunBigUInt10> a)
        {
            return (SunBigUInt10)a - (SunBigUInt10)1;
        }

        /* Relational operators */
        public override bool Greater(SunBigUIntAbstract<SunBigUInt10> a)
        {
            return SunBigUInt10.Greater((SunBigUInt10)this, (SunBigUInt10)a);
        }

        public static bool Greater(SunBigUInt10 a, SunBigUInt10 b)
        {
            if (a.digits.Length > b.digits.Length)
            {
                return true;
            }

            if (a.digits.Length < b.digits.Length)
            {
                return false;
            }

            // a and b have same number of digits
            for (int i = a.digits.Length - 1; i >= 0; i--)
            {
                if (a.digits[i] > b.digits[i])
                {
                    return true;
                }
                else if (a.digits[i] < b.digits[i])
                {
                    return false;
                }
            }

            // both values are equal, return false
            return false;
        }

        public override bool Less(SunBigUIntAbstract<SunBigUInt10> a)
        {
            return SunBigUInt10.Less((SunBigUInt10)this, (SunBigUInt10)a);
        }
        public static bool Less(SunBigUInt10 a, SunBigUInt10 b)
        {
            return !(a >= b);
        }

        public override bool GreaterEqual(SunBigUIntAbstract<SunBigUInt10> a)
        {
            return SunBigUInt10.GreaterEqual((SunBigUInt10)this, (SunBigUInt10)a);
        }
        public static bool GreaterEqual(SunBigUInt10 a, SunBigUInt10 b)
        {
            return ((a > b) || (a == b));
        }

        public override bool LessEqual(SunBigUIntAbstract<SunBigUInt10> a)
        {
            return SunBigUInt10.LessEqual((SunBigUInt10)this, (SunBigUInt10)a);
        }
        public static bool LessEqual(SunBigUInt10 a, SunBigUInt10 b)
        {
            return !(a > b);
        }

        public override bool Equal(SunBigUIntAbstract<SunBigUInt10> a)
        {
            return SunBigUInt10.Equal((SunBigUInt10)this, (SunBigUInt10)a);
        }
        public static bool Equal(SunBigUInt10 a, SunBigUInt10 b)
        {
            if (a.digits.Length != b.digits.Length)
            {
                return false;
            }
            for (int i = a.digits.Length - 1; i >= 0; i--)
            {
                if (a.digits[i] != b.digits[i])
                {
                    return false;
                }
            }
            return true;
        }

        public override bool Unequal(SunBigUIntAbstract<SunBigUInt10> a)
        {
            return SunBigUInt10.UnEqual((SunBigUInt10)this, (SunBigUInt10)a);
        }
        public static bool UnEqual(SunBigUInt10 a, SunBigUInt10 b)
        {
            return !(a == b);
        }

        /* Even / Odd */
        public override bool IsEven()
        {
            return this.digits[0] % 2 == 0;
        }
        public override bool IsOdd()
        {
            return this.digits[0] % 2 != 0;
        }


        /* Shifts */
        public static SunBigUInt10 LeftShiftByBase(SunBigUInt10 a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            byte[] temp = new byte[a.digits.Length + b];

            for (int i = 0; i < a.digits.Length; i++)
            {
                temp[i + b] = a.digits[i];
            }

            return new SunBigUInt10(temp);
        }

        public static SunBigUInt10 RightShiftByBase(SunBigUInt10 a, int b)
        {
            if (b >= a.digits.Length)
            {
                return 0;
            }

            byte[] temp = new byte[a.digits.Length - b];

            for (int i = a.digits.Length - 1; i >= b; i--)
            {
                temp[i - b] = a.digits[i];
            }

            return new SunBigUInt10(temp);
        }

        public static SunBigUInt10 LeftShiftBinary(SunBigUInt10 a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            return a * SunBigUInt10.Pow(2, b);
        }

        public static SunBigUInt10 RightShiftBinary(SunBigUInt10 a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            return a / SunBigUInt10.Pow(2, b);
        }


        /* Simple multiplication algorithm */

        public static SunBigUInt10 MulSimple(SunBigUInt10 a, SunBigUInt10 b)
        {
            SunBigUInt10 totalSum = 0;
            uint bStep = 0;
            for (int bIdx = 0; bIdx < b.digits.Length; bIdx++)
            {
                byte bDigit = b.digits[bIdx];
                byte carry = 0;

                SunBigUInt10 sumTemp = 0;
                uint aStep = 0;
                for (int aIdx = 0; aIdx < a.digits.Length; aIdx++)
                {
                    byte aDigit = a.digits[aIdx];
                    /* calculate multiplication of two single digits from a and b: a[i] * b[j] + carry */
                    byte digitMulResult = (byte)((bDigit * aDigit) + carry);

                    if (aIdx < a.digits.Length - 1)
                    {
                        /* not last multiplication of current row */
                        SunBigUInt10 digitSumShifted = SunBigUInt10.SFromLong((ulong)(digitMulResult % 10)).ShiftLeftDigits(aStep);
                        sumTemp += digitSumShifted;
                        carry = (byte)(digitMulResult / 10);
                        aStep++;
                    }
                    else
                    {
                        /* last step */
                        SunBigUInt10 digitSumShifted = SunBigUInt10.SFromLong((ulong)(digitMulResult)).ShiftLeftDigits(aStep);
                        sumTemp += digitSumShifted;
                    }
                }

                SunBigUInt10 rowSumShifted = sumTemp.ShiftLeftDigits(bStep);
                totalSum += rowSumShifted;
                bStep++;
            }
            return totalSum;
        }

        public SunBigUInt10 MulSimple(SunBigUInt10 b) { return SunBigUInt10.MulSimple(this, b); }

        public static SunBigUInt10 ShiftLeftDigits(SunBigUInt10 a, uint numOfDigits)
        {
            if (numOfDigits == 0)
            {
                return new SunBigUInt10(a.digits);
            }

            byte[] tmpArray = new byte[a.digits.Length + numOfDigits];
            for (int i = 0; i < a.digits.Length; i++)
            {
                tmpArray[i + numOfDigits] = a.digits[i];
            }

            return new SunBigUInt10(tmpArray);
        }

        public SunBigUInt10 ShiftLeftDigits(uint numOfDigits) { return SunBigUInt10.ShiftLeftDigits(this, numOfDigits); }


    }

}
