﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SunBigUInteger.SunBigUInt
{
    class SunBigUInt256 : SunBigUIntAbstract<SunBigUInt256>
    {
        private byte[] digits;

        private static readonly int BASE = 256;

        public SunBigUInt256() { }

        private SunBigUInt256(byte[] digitsTemp)
        {
            int firstPosNonZero = digitsTemp.Length - 1;
            bool isZero = true;
            for (int i = digitsTemp.Length - 1; i >= 0; i--)
            {
                if (digitsTemp[i] != 0)
                {
                    firstPosNonZero = i;
                    isZero = false;
                    break;
                }
            }

            if (isZero && (digitsTemp.Length > 1))
            {
                // all digits are zero, but have more than one element, just return valid zero element
                this.digits = new byte[1];
                this.digits[0] = 0;
                return;
            }

            this.digits = new byte[firstPosNonZero + 1];
            for (int i = 0; i <= firstPosNonZero; i++)
            {
                this.digits[i] = digitsTemp[i];
            }
        }

        public override SunBigUInt256 FromLong(ulong number)
        {
            return SunBigUInt256.SFromLong(number);
        }

        public static SunBigUInt256 SFromLong(ulong number)
        {
            if (number < 0)
            {
                throw new ArgumentOutOfRangeException("Negative numbers not supported.");
            }

            SunBigUInt256 n = new SunBigUInt256();
            if (number == 0)
            {
                n.digits = new byte[1];
                return n;
            }

            // count number of digits
            ulong numberCopy = number;
            int numDigits = 0;
            while (numberCopy > 0)
            {
                numberCopy /= (ulong)BASE;
                numDigits++;
            }

            n.digits = new byte[numDigits];
            int i = 0;
            while (number > 0)
            {
                n.digits[i] = (byte)(number % (ulong)BASE);
                number /= (ulong)BASE;
                i++;
            }

            return n;
        }

        public static SunBigUInt256 SFromString(string number)
        {
            if (String.IsNullOrEmpty(number))
            {
                throw new ArgumentOutOfRangeException("String is NULL or empty.");
            }
            // remove leading zeros but avoid empty string
            number = number.TrimStart('0');
            number = number.TrimStart(' ');
            number = number.TrimEnd(' ');
            number = number.Length > 0 ? number : "0";

            SunBigUInt256 n = 0;
            for (int i = 0; i < number.Length; i++)
            {
                char c = number[i];
                int digitVal;

                if ((c >= '0') && (c <= '9'))
                {
                    digitVal = c - '0';
                }
                else
                {
                    throw new ArgumentException("Invalid hexadecimal digit: " + c);
                }

                n = (n * (SunBigUInt256)10) + (SunBigUInt256)digitVal;
            }
            return n;

        }

        public override SunBigUInt256 FromString(string number)
        {
            return SunBigUInt256.SFromString(number);
        }

        public override SunBigUInt256 FromHexString(string number)
        {
            if (String.IsNullOrEmpty(number))
            {
                throw new ArgumentOutOfRangeException("String is NULL or empty.");
            }

            if (!number.StartsWith("0x"))
            {
                throw new ArgumentException("Hexadecimal string does not start with '0x'");
            }

            // remove leading '0x'
            number = number.Substring(2);
            // remove leading zeros but avoid empty string
            number = number.TrimStart('0');
            number = number.TrimStart(' ');
            number = number.TrimEnd(' ');
            number = number.Length > 0 ? number : "0";

            SunBigUInt256 n = 0;
            for (int i = 0; i < number.Length; i++)
            {
                char c = number[i];
                int digitVal;

                if ((c >= '0') && (c <= '9'))
                {
                    digitVal = c - '0';
                }
                else if ((c >= 'a') && (c <= 'f'))
                {
                    digitVal = c - 'a' + 10;
                }
                else if ((c >= 'A') && (c <= 'F'))
                {
                    digitVal = c - 'A' + 10;
                }
                else
                {
                    throw new ArgumentException("Invalid hexadecimal digit: " + c);
                }

                n = (n * (SunBigUInt256)16) + (SunBigUInt256)digitVal;
            }

            return n;
        }

        public static implicit operator SunBigUInt256(int number)
        {
            return SunBigUInt256.SFromLong((ulong)number);
        }

        public static implicit operator SunBigUInt256(ulong number)
        {
            return SunBigUInt256.SFromLong(number);
        }


        public override string ToString()
        {
            if (this == (SunBigUInt256)0) return "0";

            SunBigUInt256 r;
            SunBigUInt256 q = this;
            StringBuilder sb = new StringBuilder();
            while( q > (SunBigUInt256)0 )
            {
                DivRem(q, 10, out q, out r);
                sb.Insert(0, r.digits[0].ToString());
            }
            return sb.ToString();
        }

        /* Addition */
        public override SunBigUInt256 Add(SunBigUIntAbstract<SunBigUInt256> c1)
        {
            return SunBigUInt256.Add((SunBigUInt256)this, (SunBigUInt256)c1);
        }

        public static SunBigUInt256 Add(SunBigUInt256 a, SunBigUInt256 b)
        {
            byte[] temp = new byte[Math.Max(a.digits.Length, b.digits.Length) + 1];
            byte carry = 0;
            for (int i = 0; i < temp.Length; i++)
            {
                int anum = i < a.digits.Length ? a.digits[i] : 0;
                int bnum = i < b.digits.Length ? b.digits[i] : 0;
                int sumNum = anum + bnum + carry;
                if (sumNum >= BASE)
                {
                    sumNum -= BASE;
                    carry = 1;
                }
                else
                {
                    carry = 0;
                }
                temp[i] = (byte)sumNum;
            }
            return new SunBigUInt256(temp);
        }

        /* Subtraction */
        public override SunBigUInt256 Sub(SunBigUIntAbstract<SunBigUInt256> c1)
        {
            return SunBigUInt256.Sub((SunBigUInt256)this, (SunBigUInt256)c1);
        }

        public static SunBigUInt256 Sub(SunBigUInt256 a, SunBigUInt256 b)
        {
            if (b > a)
            {
                throw new OverflowException("b is greater than a.");
            }

            byte[] temp = new byte[a.digits.Length];
            byte borrow = 0;
            for (int i = 0; i < temp.Length; i++)
            {
                int anum = a.digits[i];
                int bnum = i < b.digits.Length ? b.digits[i] : 0;
                byte diffNum;
                if (anum < (bnum + borrow))
                {
                    // overflow
                    diffNum = (byte)(anum + BASE - bnum - borrow);
                    borrow = 1;
                }
                else
                {
                    diffNum = (byte)(anum - bnum - borrow);
                    borrow = 0;
                }
                temp[i] = diffNum;
            }

            return new SunBigUInt256(temp);
        }

        /* Multiplication */
        public override SunBigUInt256 Mul(SunBigUIntAbstract<SunBigUInt256> c1)
        {
            return SunBigUInt256.Mul((SunBigUInt256)this, (SunBigUInt256)c1);
        }

        public static SunBigUInt256 Mul(SunBigUInt256 a, SunBigUInt256 b)
        {
            byte[] tmpArray = new byte[a.digits.Length + b.digits.Length];

            int i, j;
            for (j = 0; j < b.digits.Length; j++)
            {
                int carry = 0;
                for (i = 0; i < a.digits.Length; i++)
                {
                    // note: max carry is 255*255 + 255 = 65280 -> fits in an int
                    carry += a.digits[i] * b.digits[j] + tmpArray[i + j];
                    tmpArray[i + j] = (byte)(carry % 256);
                    carry /= 256;
                }

                if (carry != 0)
                {
                    tmpArray[i + j] = (byte)(carry % 256);
                }
            }

            return new SunBigUInt256(tmpArray);
        }
        /* Division */
        public override SunBigUInt256 Div(SunBigUIntAbstract<SunBigUInt256> c1)
        {
            return SunBigUInt256.Div((SunBigUInt256)this, (SunBigUInt256)c1);
        }

        public static SunBigUInt256 Div(SunBigUInt256 a, SunBigUInt256 b)
        {
            SunBigUInt256 q, r;
            if (b == (SunBigUInt256)0)
            {
                throw new DivideByZeroException("b is 0");
            }

            DivRem(a, b, out q, out r);
            return q;
        }

        public override void DivRem(SunBigUIntAbstract<SunBigUInt256> a, SunBigUIntAbstract<SunBigUInt256> b, 
                                    out SunBigUIntAbstract<SunBigUInt256> q, out SunBigUIntAbstract<SunBigUInt256> r)
        {
            SunBigUInt256 q2;
            SunBigUInt256 r2;
            SunBigUInt256.DivRem((SunBigUInt256)a, (SunBigUInt256)b, out q2, out r2);
            q = q2;
            r = r2;
        }

        public override SunBigUInt256 DivRem(SunBigUIntAbstract<SunBigUInt256> a, SunBigUIntAbstract<SunBigUInt256> b, out SunBigUIntAbstract<SunBigUInt256> r)
        {
            if (b == (SunBigUInt256)0)
            {
                throw new DivideByZeroException("b is 0");
            }

            SunBigUIntAbstract<SunBigUInt256> q;
            DivRem(a, b, out q, out r);
            return (SunBigUInt256)q;
        }

        public static SunBigUInt256 DivRem(SunBigUInt256 a, SunBigUInt256 b, out SunBigUInt256 r)
        {
            if (b == (SunBigUInt256)0)
            {
                throw new DivideByZeroException("b is 0");
            }

            SunBigUInt256 q;
            DivRem(a, b, out q, out r);
            return q;
        }

        public void DivRem(SunBigUInt256 b, out SunBigUInt256 q, out SunBigUInt256 r)
        {
            SunBigUInt256.DivRem(this, b, out q, out r);
        }

        private static void DivRem(SunBigUInt256 a, SunBigUInt256 b, out SunBigUInt256 q, out SunBigUInt256 r)
        {
            if (a < b)
            {
                q = 0; r = a;
            }
            else
            {
                SunBigUInt256 qTmp, rTmp;
                DivRem(a, (SunBigUInt256)2 * b, out qTmp, out rTmp);
                if (rTmp < b)
                {
                    q = (SunBigUInt256)2 * qTmp;
                    r = rTmp;
                }
                else
                {
                    q = (SunBigUInt256)2 * qTmp + (SunBigUInt256)1;
                    r = rTmp - b;
                }
            }
        }

        /* Modulus */
        public override SunBigUInt256 Mod(SunBigUIntAbstract<SunBigUInt256> c1)
        {
            return SunBigUInt256.Mod((SunBigUInt256)this, (SunBigUInt256)c1);
        }
        public static SunBigUInt256 Mod(SunBigUInt256 a, SunBigUInt256 b)
        {
            SunBigUInt256 q, r;
            if (b == (SunBigUInt256)0)
            {
                throw new DivideByZeroException("b is 0");
            }

            DivRem(a, b, out q, out r);
            return r;
        }

        /* Pow / ModPow */
        public override SunBigUInt256 Pow(SunBigUIntAbstract<SunBigUInt256> a, SunBigUIntAbstract<SunBigUInt256> b)
        {
            return SunBigUInt256.Pow((SunBigUInt256)a, (SunBigUInt256)b);
        }
        public override SunBigUInt256 Pow(SunBigUIntAbstract<SunBigUInt256> a, int b)
        {
            return SunBigUInt256.Pow((SunBigUInt256)a, (SunBigUInt256)b);
        }
        public static SunBigUInt256 Pow(SunBigUInt256 a, SunBigUInt256 b)
        {
            SunBigUInt256 res = 1;
            while (b > (SunBigUInt256)0)
            {
                if (b.IsOdd())
                {
                    res *= a;
                }
                b /= (SunBigUInt256)2;
                a = a * a;
            }

            return res;
        }

        public override SunBigUInt256 ModPow(SunBigUIntAbstract<SunBigUInt256> value, SunBigUIntAbstract<SunBigUInt256> exponent, SunBigUIntAbstract<SunBigUInt256> modulus)
        {
            return SunBigUInt256.ModPow((SunBigUInt256)value, (SunBigUInt256)exponent, (SunBigUInt256)modulus);
        }
        public override SunBigUInt256 ModPow(int value, int exponent, int modulus)
        {
            return SunBigUInt256.ModPow((SunBigUInt256)value, (SunBigUInt256)exponent, (SunBigUInt256)modulus);
        }
        public static SunBigUInt256 ModPow(SunBigUInt256 value, SunBigUInt256 exponent, SunBigUInt256 modulus)
        {
            if (modulus == (SunBigUInt256)1)
            {
                return 0;
            }

            SunBigUInt256 c = 1;
            for (SunBigUInt256 i = 0; i < exponent; i++)
            {
                c = (c * value).Mod(modulus);
            }
            return c;
        }

        /* Increment / Decrement */
        public override SunBigUInt256 Inc(SunBigUIntAbstract<SunBigUInt256> a)
        {
            return (SunBigUInt256)a + (SunBigUInt256)1;
        }
        public override SunBigUInt256 Dec(SunBigUIntAbstract<SunBigUInt256> a)
        {
            return (SunBigUInt256)a - (SunBigUInt256)1;
        }

        /* Relational operators */
        public override bool Greater(SunBigUIntAbstract<SunBigUInt256> a)
        {
            return SunBigUInt256.Greater((SunBigUInt256)this, (SunBigUInt256)a);
        }

        public static bool Greater(SunBigUInt256 a, SunBigUInt256 b)
        {
            if (a.digits.Length > b.digits.Length)
            {
                return true;
            }

            if (a.digits.Length < b.digits.Length)
            {
                return false;
            }

            // a and b have same number of digits
            for (int i = a.digits.Length - 1; i >= 0; i--)
            {
                if (a.digits[i] > b.digits[i])
                {
                    return true;
                }
                else if (a.digits[i] < b.digits[i])
                {
                    return false;
                }
            }

            // both values are equal, return false
            return false;
        }

        public override bool Less(SunBigUIntAbstract<SunBigUInt256> a)
        {
            return SunBigUInt256.Less((SunBigUInt256)this, (SunBigUInt256)a);
        }
        public static bool Less(SunBigUInt256 a, SunBigUInt256 b)
        {
            return !(a >= b);
        }

        public override bool GreaterEqual(SunBigUIntAbstract<SunBigUInt256> a)
        {
            return SunBigUInt256.GreaterEqual((SunBigUInt256)this, (SunBigUInt256)a);
        }
        public static bool GreaterEqual(SunBigUInt256 a, SunBigUInt256 b)
        {
            return ((a > b) || (a == b));
        }

        public override bool LessEqual(SunBigUIntAbstract<SunBigUInt256> a)
        {
            return SunBigUInt256.LessEqual((SunBigUInt256)this, (SunBigUInt256)a);
        }
        public static bool LessEqual(SunBigUInt256 a, SunBigUInt256 b)
        {
            return !(a > b);
        }

        public override bool Equal(SunBigUIntAbstract<SunBigUInt256> a)
        {
            return SunBigUInt256.Equal((SunBigUInt256)this, (SunBigUInt256)a);
        }
        public static bool Equal(SunBigUInt256 a, SunBigUInt256 b)
        {
            if (a.digits.Length != b.digits.Length)
            {
                return false;
            }
            for (int i = a.digits.Length - 1; i >= 0; i--)
            {
                if (a.digits[i] != b.digits[i])
                {
                    return false;
                }
            }
            return true;
        }

        public override bool Unequal(SunBigUIntAbstract<SunBigUInt256> a)
        {
            return SunBigUInt256.UnEqual((SunBigUInt256)this, (SunBigUInt256)a);
        }
        public static bool UnEqual(SunBigUInt256 a, SunBigUInt256 b)
        {
            return !(a == b);
        }

        /* Even / Odd */
        public override bool IsEven()
        {
            return this.digits[0] % 2 == 0;
        }
        public override bool IsOdd()
        {
            return this.digits[0] % 2 != 0;
        }


        /* Shifts */
        public static SunBigUInt256 LeftShiftByBase(SunBigUInt256 a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            byte[] temp = new byte[a.digits.Length + b];

            for (int i = 0; i < a.digits.Length; i++)
            {
                temp[i + b] = a.digits[i];
            }

            return new SunBigUInt256(temp);
        }

        public static SunBigUInt256 RightShiftByBase(SunBigUInt256 a, int b)
        {
            if (b >= a.digits.Length)
            {
                return 0;
            }

            byte[] temp = new byte[a.digits.Length - b];

            for (int i = a.digits.Length - 1; i >= b; i--)
            {
                temp[i - b] = a.digits[i];
            }

            return new SunBigUInt256(temp);
        }

        public static SunBigUInt256 LeftShiftBinary(SunBigUInt256 a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            return a * SunBigUInt256.Pow(2, b);
        }

        public static SunBigUInt256 RightShiftBinary(SunBigUInt256 a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            return a / SunBigUInt256.Pow(2, b);
        }
    }
}
