﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SunBigUInteger.SunBigUIntConcrete
{
    class SunBigUInt10Concrete
    {
        private byte[] digits;

        private SunBigUInt10Concrete() { }

        private SunBigUInt10Concrete(byte[] digitsTemp)
        {
            int firstPosNonZero = digitsTemp.Length - 1;
            bool isZero = true;
            for (int i = digitsTemp.Length - 1; i >= 0; i--)
            {
                if (digitsTemp[i] != 0)
                {
                    firstPosNonZero = i;
                    isZero = false;
                    break;
                }
            }

            if (isZero && (digitsTemp.Length > 1))
            {
                // all digits are zero, but have more than one element, just return valid zero element
                this.digits = new byte[1];
                this.digits[0] = 0;
                return;
            }

            this.digits = new byte[firstPosNonZero + 1];
            for (int i = 0; i <= firstPosNonZero; i++)
            {
                this.digits[i] = digitsTemp[i];
            }
        }

        public static SunBigUInt10Concrete FromLong(ulong number)
        {
            if (number < 0)
            {
                throw new ArgumentOutOfRangeException("Negative numbers not supported.");
            }

            SunBigUInt10Concrete n = new SunBigUInt10Concrete();
            if (number == 0)
            {
                n.digits = new byte[1];
                return n;
            }

            // count number of digits
            ulong numberCopy = number;
            int numDigits = 0;
            while (numberCopy > 0)
            {
                numberCopy /= 10;
                numDigits++;
            }

            n.digits = new byte[numDigits];
            int i = 0;
            while (number > 0)
            {
                n.digits[i] = (byte)(number % 10);
                number /= 10;
                i++;
            }

            return n;
        }

        public static SunBigUInt10Concrete FromString(string number)
        {
            if (String.IsNullOrEmpty(number))
            {
                throw new ArgumentOutOfRangeException("String is NULL or empty.");
            }
            // remove leading zeros but avoid empty string
            number = number.TrimStart('0');
            number = number.TrimStart(' ');
            number = number.TrimEnd(' ');
            number = number.Length > 0 ? number : "0";

            SunBigUInt10Concrete n = new SunBigUInt10Concrete();
            n.digits = new byte[number.Length];

            for (int i = 0; i < number.Length; i++)
            {
                if (!Char.IsDigit(number[number.Length - i - 1]))
                {
                    throw new ArgumentException("Invalid decimal digit: " + number[number.Length - i - 1]);
                }
                n.digits[i] = (byte)(number[number.Length - i - 1] - 0x30);
            }
            return n;
        }

        public static SunBigUInt10Concrete FromHexString(string number)
        {
            if (String.IsNullOrEmpty(number))
            {
                throw new ArgumentOutOfRangeException("String is NULL or empty.");
            }

            if (!number.StartsWith("0x"))
            {
                throw new ArgumentException("Hexadecimal string does not start with '0x'");
            }

            // remove leading '0x'
            number = number.Substring(2);
            // remove leading zeros but avoid empty string
            number = number.TrimStart('0');
            number = number.TrimStart(' ');
            number = number.TrimEnd(' ');
            number = number.Length > 0 ? number : "0";

            SunBigUInt10Concrete n = 0;
            for (int i = 0; i < number.Length; i++)
            {
                char c = number[i];
                int digitVal;

                if ((c >= '0') && (c <= '9'))
                {
                    digitVal = c - '0';
                }
                else if ((c >= 'a') && (c <= 'f'))
                {
                    digitVal = c - 'a' + 10;
                }
                else if ((c >= 'A') && (c <= 'F'))
                {
                    digitVal = c - 'A' + 10;
                }
                else
                {
                    throw new ArgumentException("Invalid hexadecimal digit: " + c);
                }

                n = (n * 16) + digitVal;
            }

            return n;
        }

        public static implicit operator SunBigUInt10Concrete(int number)
        {
            return SunBigUInt10Concrete.FromLong((ulong)number);
        }

        public static implicit operator SunBigUInt10Concrete(ulong number)
        {
            return SunBigUInt10Concrete.FromLong(number);
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            for (int i = this.digits.Length - 1; i >= 0; i--)
            {
                sb.Append(digits[i].ToString());
            }
            return sb.ToString();
        }

        public static SunBigUInt10Concrete operator +(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            byte[] temp = new byte[Math.Max(a.digits.Length, b.digits.Length) + 1];
            byte carry = 0;
            for (int i = 0; i < temp.Length; i++)
            {
                int anum = i < a.digits.Length ? a.digits[i] : 0;
                int bnum = i < b.digits.Length ? b.digits[i] : 0;
                byte sumNum = (byte)(anum + bnum + carry);
                if (sumNum >= 10)
                {
                    sumNum -= 10;
                    carry = 1;
                }
                else
                {
                    carry = 0;
                }
                temp[i] = sumNum;
            }
            return new SunBigUInt10Concrete(temp);
        }

        public static SunBigUInt10Concrete operator -(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            if (b > a)
            {
                throw new OverflowException("b is greater than a.");
            }

            byte[] temp = new byte[a.digits.Length];
            byte borrow = 0;
            for (int i = 0; i < temp.Length; i++)
            {
                int anum = a.digits[i];
                int bnum = i < b.digits.Length ? b.digits[i] : 0;
                byte diffNum;
                if (anum < (bnum + borrow))
                {
                    // overflow
                    diffNum = (byte)(anum + 10 - bnum - borrow);
                    borrow = 1;
                }
                else
                {
                    diffNum = (byte)(anum - bnum - borrow);
                    borrow = 0;
                }
                temp[i] = diffNum;
            }

            return new SunBigUInt10Concrete(temp);
        }

        public static SunBigUInt10Concrete MulSimple(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            SunBigUInt10Concrete result = 0;
            uint bStep = 0;
            for (int bIdx = 0; bIdx < b.digits.Length; bIdx++)
            {
                byte bDigit = b.digits[bIdx];
                byte carry = 0;

                SunBigUInt10Concrete rowProduct = 0;
                uint aStep = 0;
                for (int aIdx = 0; aIdx < a.digits.Length; aIdx++)
                {
                    byte aDigit = a.digits[aIdx];
                    /* calculate multiplication of two single digits from a and b: a[i] * b[j] + carry */
                    byte digitMulResult = (byte)((bDigit * aDigit) + carry);

                    if (aIdx < a.digits.Length - 1)
                    {
                        /* not last multiplication of current row, store digit in sum of current row, update carry */
                        SunBigUInt10Concrete digitSumShifted = SunBigUInt10Concrete.FromLong((ulong)(digitMulResult % 10)).ShiftLeftDigits(aStep);
                        rowProduct += digitSumShifted;
                        carry = (byte)(digitMulResult / 10);
                        aStep++;
                    }
                    else
                    {
                        /* last step takes into account that no leading zero is appended */
                        SunBigUInt10Concrete digitSumShifted = SunBigUInt10Concrete.FromLong((ulong)(digitMulResult)).ShiftLeftDigits(aStep);
                        rowProduct += digitSumShifted;
                    }
                }
                /* multipy sum of row n with 10^n by left shifting by n */
                SunBigUInt10Concrete rowProductSumShifted = rowProduct.ShiftLeftDigits(bStep);
                /* add rowproduct to total result */
                result += rowProductSumShifted;
                bStep++;
            }
            return result;
        }

        public SunBigUInt10Concrete MulSimple(SunBigUInt10Concrete b) { return SunBigUInt10Concrete.MulSimple(this, b); }



        public static SunBigUInt10Concrete ShiftLeftDigits(SunBigUInt10Concrete a, uint numOfDigits)
        {
            if (numOfDigits == 0)
            {
                return new SunBigUInt10Concrete(a.digits);
            }

            byte[] tmpArray = new byte[a.digits.Length + numOfDigits];
            for (int i = 0; i < a.digits.Length; i++)
            {
                tmpArray[i + numOfDigits] = a.digits[i];
            }

            return new SunBigUInt10Concrete(tmpArray);
        }

        public SunBigUInt10Concrete ShiftLeftDigits(uint numOfDigits) { return SunBigUInt10Concrete.ShiftLeftDigits(this, numOfDigits);  }

        public static SunBigUInt10Concrete Mul(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            byte[] tmpArray = new byte[a.digits.Length + b.digits.Length];

            int i, j;
            for (j = 0; j < b.digits.Length; j++)
            {
                byte carry = 0;
                for (i = 0; i < a.digits.Length; i++)
                {
                    // note: max carry is 9*9 + 9 = 90 -> fits in a byte
                    carry += (byte)(a.digits[i] * b.digits[j] + tmpArray[i + j]);
                    tmpArray[i + j] = (byte)(carry % 10);
                    carry /= 10;
                }

                if (carry != 0)
                {
                    tmpArray[i + j] = (byte)(carry % 10);
                }
            }

            return new SunBigUInt10Concrete(tmpArray);
        }

        public SunBigUInt10Concrete Mul(SunBigUInt10Concrete b) { return SunBigUInt10Concrete.Mul(this, b); }

        public static SunBigUInt10Concrete operator *(SunBigUInt10Concrete a, SunBigUInt10Concrete b) { return SunBigUInt10Concrete.Mul(a, b); }

        private static void DivRem(SunBigUInt10Concrete a, SunBigUInt10Concrete b, out SunBigUInt10Concrete q, out SunBigUInt10Concrete r)
        {
            if (a < b)
            {
                q = 0; r = a;
            }
            else
            {
                SunBigUInt10Concrete qTmp, rTmp;
                DivRem(a, 2 * b, out qTmp, out rTmp);
                if (rTmp < b)
                {
                    q = 2 * qTmp;
                    r = rTmp;
                }
                else
                {
                    q = 2 * qTmp + 1;
                    r = rTmp - b;
                }
            }
        }

        public static SunBigUInt10Concrete DivRem(SunBigUInt10Concrete a, SunBigUInt10Concrete b, out SunBigUInt10Concrete r)
        {
            if (b == 0)
            {
                throw new DivideByZeroException("b is 0");
            }

            SunBigUInt10Concrete q;
            DivRem(a, b, out q, out r);
            return q;
        }

        public void DivRem(SunBigUInt10Concrete b, out SunBigUInt10Concrete q, out SunBigUInt10Concrete r)
        {
            SunBigUInt10Concrete.DivRem(this, b, out q, out r);
        }

        public static SunBigUInt10Concrete Div(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            SunBigUInt10Concrete q, r;
            if (b == 0)
            {
                throw new DivideByZeroException("b is 0");
            }

            DivRem(a, b, out q, out r);
            return q;
        }

        public SunBigUInt10Concrete Div(SunBigUInt10Concrete b) 
        {
            return SunBigUInt10Concrete.Div(this, b);
        }

        public static SunBigUInt10Concrete operator /(SunBigUInt10Concrete a, SunBigUInt10Concrete b) { return SunBigUInt10Concrete.Div(a, b); }

        public static SunBigUInt10Concrete Mod(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            SunBigUInt10Concrete q, r;
            if (b == 0)
            {
                throw new DivideByZeroException("b is 0");
            }

            DivRem(a, b, out q, out r);
            return r;
        }

        public SunBigUInt10Concrete Mod(SunBigUInt10Concrete b)
        {
            return SunBigUInt10Concrete.Mod(this, b);
        }

        public static SunBigUInt10Concrete Pow(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            SunBigUInt10Concrete res = 1;
            while (b > 0)
            {
                if (b.IsOdd())
                {
                    res *= a;
                }
                b /= 2;
                a = a * a;
            }

            return res;
        }

        public static SunBigUInt10Concrete ModPow(SunBigUInt10Concrete value, SunBigUInt10Concrete exponent, SunBigUInt10Concrete modulus)
        {
            if (modulus == 1)
            {
                return 0;
            }

            SunBigUInt10Concrete c = 1;
            for (SunBigUInt10Concrete i = 0; i < exponent; i++)
            {
                c = (c * value).Mod(modulus);
            }
            return c;
        }

        public static SunBigUInt10Concrete LeftShiftByBase(SunBigUInt10Concrete a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            byte[] temp = new byte[a.digits.Length + b];

            for (int i = 0; i < a.digits.Length; i++)
            {
                temp[i + b] = a.digits[i];
            }

            return new SunBigUInt10Concrete(temp);
        }

        public static SunBigUInt10Concrete RightShiftByBase(SunBigUInt10Concrete a, int b)
        {
            if (b >= a.digits.Length)
            {
                return 0;
            }

            byte[] temp = new byte[a.digits.Length - b];

            for (int i = a.digits.Length - 1 ; i >= b; i--)
            {
                temp[i - b] = a.digits[i];
            }

            return new SunBigUInt10Concrete(temp);
        }

        public static SunBigUInt10Concrete LeftShiftBinary(SunBigUInt10Concrete a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            return a * SunBigUInt10Concrete.Pow(2, b);
        }

        public static SunBigUInt10Concrete RightShiftBinary(SunBigUInt10Concrete a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            return a / SunBigUInt10Concrete.Pow(2, b);
        }

        public static SunBigUInt10Concrete operator ++(SunBigUInt10Concrete a)
        {
            return a + 1;
        }

        public static SunBigUInt10Concrete operator --(SunBigUInt10Concrete a)
        {
            return a - 1;
        }

        public static bool operator >(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            if (a.digits.Length > b.digits.Length)
            {
                return true;
            }

            if (a.digits.Length < b.digits.Length)
            {
                return false;
            }

            // a and b have same number of digits
            for (int i = a.digits.Length - 1; i >= 0; i--)
            {
                if (a.digits[i] > b.digits[i])
                {
                    return true;
                }
                else if (a.digits[i] < b.digits[i])
                {
                    return false;
                }
            }

            // both values are equal, return false
            return false;
        }

        public static bool operator >=(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            return ((a > b) || (a == b));
        }

        public static bool operator <(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            return !(a >= b);
        }

        public static bool operator <=(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            return !(a > b);
        }

        public static bool operator ==(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            if (a.digits.Length != b.digits.Length)
            {
                return false;
            }
            for (int i = a.digits.Length - 1; i >= 0; i--)
            {
                if (a.digits[i] != b.digits[i])
                {
                    return false;
                }
            }
            return true;
        }

        public static bool operator !=(SunBigUInt10Concrete a, SunBigUInt10Concrete b)
        {
            return !(a == b);
        }

        public bool IsEven()
        {
            return this.digits[0] % 2 == 0;
        }

        public bool IsOdd()
        {
            return this.digits[0] % 2 != 0;
        }

        public override bool Equals(object obj)
        {
            if (obj.GetType() != this.GetType())
            {
                return false;
            }

            return this.Equals(obj as SunBigUInt10Concrete);
        }

        public bool Equals(SunBigUInt10Concrete obj)
        {
            if (obj is null)
            {
                return false;
            }

            if (Object.ReferenceEquals(this, obj))
            {
                return true;
            }

            return this == obj;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

    }
}
