﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SunBigUInteger.SunBigUIntConcrete
{
    class SunBigUInt256Concrete
    {
        private byte[] digits;

        private static readonly int BASE = 256;

        private SunBigUInt256Concrete() { }

        private SunBigUInt256Concrete(byte[] digitsTemp)
        {
            int firstPosNonZero = digitsTemp.Length - 1;
            bool isZero = true;
            for (int i = digitsTemp.Length - 1; i >= 0; i--)
            {
                if (digitsTemp[i] != 0)
                {
                    firstPosNonZero = i;
                    isZero = false;
                    break;
                }
            }

            if (isZero && (digitsTemp.Length > 1))
            {
                // all digits are zero, but have more than one element, just return valid zero element
                this.digits = new byte[1];
                this.digits[0] = 0;
                return;
            }

            this.digits = new byte[firstPosNonZero + 1];
            for (int i = 0; i <= firstPosNonZero; i++)
            {
                this.digits[i] = digitsTemp[i];
            }
        }

        public static SunBigUInt256Concrete FromLong(ulong number)
        {
            if (number < 0)
            {
                throw new ArgumentOutOfRangeException("Negative numbers not supported.");
            }

            SunBigUInt256Concrete n = new SunBigUInt256Concrete();
            if (number == 0)
            {
                n.digits = new byte[1];
                return n;
            }

            // count number of digits
            ulong numberCopy = number;
            int numDigits = 0;
            while (numberCopy > 0)
            {
                numberCopy /= (ulong)BASE;
                numDigits++;
            }

            n.digits = new byte[numDigits];
            int i = 0;
            while (number > 0)
            {
                n.digits[i] = (byte)(number % (ulong)BASE);
                number /= (ulong)BASE;
                i++;
            }

            return n;
        }

        public static SunBigUInt256Concrete FromString(string number)
        {
            if (String.IsNullOrEmpty(number))
            {
                throw new ArgumentOutOfRangeException("String is NULL or empty.");
            }
            // remove leading zeros but avoid empty string
            number = number.TrimStart('0');
            number = number.TrimStart(' ');
            number = number.TrimEnd(' ');
            number = number.Length > 0 ? number : "0";

            SunBigUInt256Concrete n = 0;
            for (int i = 0; i < number.Length; i++)
            {
                char c = number[i];
                int digitVal;

                if ((c >= '0') && (c <= '9'))
                {
                    digitVal = c - '0';
                }
                else
                {
                    throw new ArgumentException("Invalid hexadecimal digit: " + c);
                }

                n = (n * 10) + digitVal;
            }
            return n;
        }

        public static SunBigUInt256Concrete FromHexString(string number)
        {
            if (String.IsNullOrEmpty(number))
            {
                throw new ArgumentOutOfRangeException("String is NULL or empty.");
            }

            if (!number.StartsWith("0x"))
            {
                throw new ArgumentException("Hexadecimal string does not start with '0x'");
            }

            // remove leading '0x'
            number = number.Substring(2);
            // remove leading zeros but avoid empty string
            number = number.TrimStart('0');
            number = number.TrimStart(' ');
            number = number.TrimEnd(' ');
            number = number.Length > 0 ? number : "0";

            SunBigUInt256Concrete n = 0;
            for (int i = 0; i < number.Length; i++)
            {
                char c = number[i];
                int digitVal;

                if ((c >= '0') && (c <= '9'))
                {
                    digitVal = c - '0';
                }
                else if ((c >= 'a') && (c <= 'f'))
                {
                    digitVal = c - 'a' + 10;
                }
                else if ((c >= 'A') && (c <= 'F'))
                {
                    digitVal = c - 'A' + 10;
                }
                else
                {
                    throw new ArgumentException("Invalid hexadecimal digit: " + c);
                }

                n = (n * 16) + digitVal;
            }

            return n;
        }

        public static implicit operator SunBigUInt256Concrete(int number)
        {
            return SunBigUInt256Concrete.FromLong((ulong)number);
        }

        public static implicit operator SunBigUInt256Concrete(ulong number)
        {
            return SunBigUInt256Concrete.FromLong(number);
        }

        public override string ToString()
        {
            if (this == 0) return "0";

            SunBigUInt256Concrete r;
            SunBigUInt256Concrete q = this;
            StringBuilder sb = new StringBuilder();
            while( q > 0 )
            {
                DivRem(q, 10, out q, out r);
                sb.Insert(0, r.digits[0].ToString());
            }
            return sb.ToString();
        }

        public static SunBigUInt256Concrete operator +(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            byte[] temp = new byte[Math.Max(a.digits.Length, b.digits.Length) + 1];
            byte carry = 0;
            for (int i = 0; i < temp.Length; i++)
            {
                int anum = i < a.digits.Length ? a.digits[i] : 0;
                int bnum = i < b.digits.Length ? b.digits[i] : 0;
                int sumNum = anum + bnum + carry;
                if (sumNum >= BASE)
                {
                    sumNum -= BASE;
                    carry = 1;
                }
                else
                {
                    carry = 0;
                }
                temp[i] = (byte)sumNum;
            }
            return new SunBigUInt256Concrete(temp);
        }

        public static SunBigUInt256Concrete operator -(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            if (b > a)
            {
                throw new OverflowException("b is greater than a.");
            }

            byte[] temp = new byte[a.digits.Length];
            byte borrow = 0;
            for (int i = 0; i < temp.Length; i++)
            {
                int anum = a.digits[i];
                int bnum = i < b.digits.Length ? b.digits[i] : 0;
                byte diffNum;
                if (anum < (bnum + borrow))
                {
                    // overflow
                    diffNum = (byte)(anum + BASE - bnum - borrow);
                    borrow = 1;
                }
                else
                {
                    diffNum = (byte)(anum - bnum - borrow);
                    borrow = 0;
                }
                temp[i] = diffNum;
            }

            return new SunBigUInt256Concrete(temp);
        }

        public static SunBigUInt256Concrete Mul(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            byte[] tmpArray = new byte[a.digits.Length + b.digits.Length];

            int i, j;
            for (j = 0; j < b.digits.Length; j++)
            {
                int carry = 0;
                for (i = 0; i < a.digits.Length; i++)
                {
                    // note: max carry is 255*255 + 255 = 65280 -> fits in an int
                    carry += a.digits[i] * b.digits[j] + tmpArray[i + j];
                    tmpArray[i + j] = (byte)(carry % 256);
                    carry /= 256;
                }

                if (carry != 0)
                {
                    tmpArray[i + j] = (byte)(carry % 256);
                }
            }

            return new SunBigUInt256Concrete(tmpArray);
        }

        public SunBigUInt256Concrete Mul(SunBigUInt256Concrete b) { return SunBigUInt256Concrete.Mul(this, b); }

        public static SunBigUInt256Concrete operator *(SunBigUInt256Concrete a, SunBigUInt256Concrete b) { return SunBigUInt256Concrete.Mul(a, b); }


        private static void DivRem(SunBigUInt256Concrete a, SunBigUInt256Concrete b, out SunBigUInt256Concrete q, out SunBigUInt256Concrete r)
        {
            if (a < b)
            {
                q = 0; r = a;
            }
            else
            {
                SunBigUInt256Concrete qTmp, rTmp;
                DivRem(a, 2 * b, out qTmp, out rTmp);
                if (rTmp < b)
                {
                    q = 2 * qTmp;
                    r = rTmp;
                }
                else
                {
                    q = 2 * qTmp + 1;
                    r = rTmp - b;
                }
            }
        }

        public static SunBigUInt256Concrete DivRem(SunBigUInt256Concrete a, SunBigUInt256Concrete b, out SunBigUInt256Concrete r)
        {
            if (b == 0)
            {
                throw new DivideByZeroException("b is 0");
            }

            SunBigUInt256Concrete q;
            DivRem(a, b, out q, out r);
            return q;
        }

        public void DivRem(SunBigUInt256Concrete b, out SunBigUInt256Concrete q, out SunBigUInt256Concrete r)
        {
            SunBigUInt256Concrete.DivRem(this, b, out q, out r);
        }

        public static SunBigUInt256Concrete Div(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            SunBigUInt256Concrete q, r;
            if (b == 0)
            {
                throw new DivideByZeroException("b is 0");
            }

            DivRem(a, b, out q, out r);
            return q;
        }

        public SunBigUInt256Concrete Div(SunBigUInt256Concrete b)
        {
            return SunBigUInt256Concrete.Div(this, b);
        }

        public static SunBigUInt256Concrete operator /(SunBigUInt256Concrete a, SunBigUInt256Concrete b) { return SunBigUInt256Concrete.Div(a, b); }

        public static SunBigUInt256Concrete Mod(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            SunBigUInt256Concrete q, r;
            if (b == 0)
            {
                throw new DivideByZeroException("b is 0");
            }

            DivRem(a, b, out q, out r);
            return r;
        }

        public SunBigUInt256Concrete Mod(SunBigUInt256Concrete b)
        {
            return SunBigUInt256Concrete.Mod(this, b);
        }

        public static SunBigUInt256Concrete Pow(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            SunBigUInt256Concrete res = 1;
            while (b > 0)
            {
                if (b.IsOdd())
                {
                    res *= a;
                }
                b /= 2;
                a = a * a;
            }

            return res;
        }

        public static SunBigUInt256Concrete ModPow(SunBigUInt256Concrete value, SunBigUInt256Concrete exponent, SunBigUInt256Concrete modulus)
        {
            if (modulus == 1)
            {
                return 0;
            }

            SunBigUInt256Concrete c = 1;
            for (SunBigUInt256Concrete i = 0; i < exponent; i++)
            {
                c = (c * value).Mod(modulus);
            }
            return c;
        }

        public static SunBigUInt256Concrete LeftShiftByBase(SunBigUInt256Concrete a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            byte[] temp = new byte[a.digits.Length + b];

            for (int i = 0; i < a.digits.Length; i++)
            {
                temp[i + b] = a.digits[i];
            }

            return new SunBigUInt256Concrete(temp);
        }

        public static SunBigUInt256Concrete RightShiftByBase(SunBigUInt256Concrete a, int b)
        {
            if (b >= a.digits.Length)
            {
                return 0;
            }

            byte[] temp = new byte[a.digits.Length - b];

            for (int i = a.digits.Length - 1; i >= b; i--)
            {
                temp[i - b] = a.digits[i];
            }

            return new SunBigUInt256Concrete(temp);
        }

        public static SunBigUInt256Concrete LeftShiftBinary(SunBigUInt256Concrete a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            return a * SunBigUInt256Concrete.Pow(2, b);
        }

        public static SunBigUInt256Concrete RightShiftBinary(SunBigUInt256Concrete a, int b)
        {
            if (b < 0)
            {
                throw new ArgumentOutOfRangeException("b must not be negative");
            }

            return a / SunBigUInt256Concrete.Pow(2, b);
        }

        public static SunBigUInt256Concrete operator ++(SunBigUInt256Concrete a)
        {
            return a + 1;
        }

        public static SunBigUInt256Concrete operator --(SunBigUInt256Concrete a)
        {
            return a - 1;
        }

        public static bool operator >(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            if (a.digits.Length > b.digits.Length)
            {
                return true;
            }

            if (a.digits.Length < b.digits.Length)
            {
                return false;
            }

            // a and b have same number of digits
            for (int i = a.digits.Length - 1; i >= 0; i--)
            {
                if (a.digits[i] > b.digits[i])
                {
                    return true;
                }
                else if (a.digits[i] < b.digits[i])
                {
                    return false;
                }
            }

            // both values are equal, return false
            return false;
        }

        public static bool operator >=(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            return ((a > b) || (a == b));
        }

        public static bool operator <(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            return !(a >= b);
        }

        public static bool operator <=(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            return !(a > b);
        }

        public static bool operator ==(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            if (a.digits.Length != b.digits.Length)
            {
                return false;
            }
            for (int i = a.digits.Length - 1; i >= 0; i--)
            {
                if (a.digits[i] != b.digits[i])
                {
                    return false;
                }
            }
            return true;
        }

        public static bool operator !=(SunBigUInt256Concrete a, SunBigUInt256Concrete b)
        {
            return !(a == b);
        }

        public bool IsEven()
        {
            return this.digits[0] % 2 == 0;
        }

        public bool IsOdd()
        {
            return this.digits[0] % 2 != 0;
        }

        public override bool Equals(object obj)
        {
            if (obj.GetType() != this.GetType())
            {
                return false;
            }

            return this.Equals(obj as SunBigUInt256Concrete);
        }

        public bool Equals(SunBigUInt256Concrete obj)
        {
            if (obj is null)
            {
                return false;
            }

            if (Object.ReferenceEquals(this, obj))
            {
                return true;
            }

            return this == obj;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }
}
