﻿using SunBigUInteger.SunBigUInt;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Text;
using System.Threading.Tasks;

namespace SunBigUInteger.SunBigUIntTests
{
    class SunBigUIntTestSuite<TClass> where TClass : SunBigUIntAbstract<TClass>
    {
        private readonly ITestHandler<TClass> testHandler;
        public int NumOfFailedTestcases { private set; get; }
        public enum COMPARISON_OPERATOR
        {
            GREATER,
            GREATERTHAN,
            SMALLER,
            SMALLERTHAN,
            EQUALITY,
            INEQUALITY
        }

        public SunBigUIntTestSuite(ITestHandler<TClass> ts) 
        {
            this.testHandler = ts;
            NumOfFailedTestcases = 0;
        }

        public void Run<T>(Type type) where T : SunBigUIntAbstract<T>
        {
            T sunBigInt = (T)Activator.CreateInstance(type);

            RunConstructorNumericalTestcases(sunBigInt);
            RunConstructorStringTestcases(sunBigInt);
            RunConstructorHexStringTestcases(sunBigInt);

            RunAdditionTestcases(sunBigInt);
            RunSubtractionTestcases(sunBigInt);

            RunGreaterTestCases(sunBigInt);
            RunGreaterThanTestCases(sunBigInt);
            RunLessTestCases(sunBigInt);
            RunLessThanTestCases(sunBigInt);
            RunEqualityThanTestCases(sunBigInt);
            RunInequalityThanTestCases(sunBigInt);

            RunIsEvenTestCases(sunBigInt);
            RunIsOddTestCases(sunBigInt);

            RunMultplicationTestCases(sunBigInt);
            RunDivisionTestCases(sunBigInt);
            RunPowTestCases(sunBigInt);
            RunModPowTestCases(sunBigInt);
        }

        public void RunConstructorNumericalTestcases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Constructor (numerical)");

            AssertEqual(new BigInteger(123), sunBigInt.FromLong(123));
            AssertEqual(new BigInteger(233), sunBigInt.FromLong(233));
            AssertEqual(new BigInteger(0), sunBigInt.FromLong(0));
            AssertEqual(new BigInteger(1), sunBigInt.FromLong(1));
            AssertEqual(new BigInteger(123456789333), sunBigInt.FromLong(123456789333));
            AssertEqual(new BigInteger(9981238843545789333), sunBigInt.FromLong(9981238843545789333));

            testHandler.TestGroupEnd();
        }

        public void RunConstructorStringTestcases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Constructor (string)");

            AssertEqual(BigInteger.Parse("0"), sunBigInt.FromString("0"));
            AssertEqual(BigInteger.Parse("11"), sunBigInt.FromString("11"));
            AssertEqual(BigInteger.Parse("00002"), sunBigInt.FromString("00002"));
            AssertEqual(BigInteger.Parse("11111111111111111111111111111111111110"), sunBigInt.FromString("11111111111111111111111111111111111110"));
            AssertEqual(BigInteger.Parse("33445566778899112234567890000001"), sunBigInt.FromString("33445566778899112234567890000001"));
            AssertEqual(BigInteger.Parse("9999999999999999999"), sunBigInt.FromString("9999999999999999999"));
            AssertEqual(BigInteger.Parse(" 39986 "), sunBigInt.FromString(" 39986 "));

            AssertException<ArgumentException>(() => sunBigInt.FromString(""));
            AssertException<ArgumentException>(() => sunBigInt.FromString("-1"));
            AssertException<ArgumentException>(() => sunBigInt.FromString("121333a"));
            AssertException<ArgumentException>(() => sunBigInt.FromString(".12222"));

            testHandler.TestGroupEnd();
        }

        public void RunConstructorHexStringTestcases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Constructor (Hex string)");
            AssertEqual(BigInteger.Parse("0", System.Globalization.NumberStyles.HexNumber), sunBigInt.FromHexString("0x0"));
            AssertEqual(BigInteger.Parse("0", System.Globalization.NumberStyles.HexNumber), sunBigInt.FromHexString("0x00"));
            AssertEqual(BigInteger.Parse("0", System.Globalization.NumberStyles.HexNumber), sunBigInt.FromHexString("0x000"));
            AssertEqual(BigInteger.Parse("0", System.Globalization.NumberStyles.HexNumber), sunBigInt.FromHexString("0x0000"));
            AssertEqual(BigInteger.Parse("11", System.Globalization.NumberStyles.HexNumber), sunBigInt.FromHexString("0x11"));
            AssertEqual(BigInteger.Parse("11", System.Globalization.NumberStyles.HexNumber), sunBigInt.FromHexString("0x00011"));
            /* Note: Add a leading '0' to the BigInteger numbers to tell it it's positive */
            AssertEqual(BigInteger.Parse("0A", System.Globalization.NumberStyles.HexNumber), sunBigInt.FromHexString("0xA"));
            AssertEqual(14, sunBigInt.FromHexString("0x0e"));
            AssertEqual(14, sunBigInt.FromHexString("0xe"));
            AssertEqual(14, sunBigInt.FromHexString("0xE"));
            AssertEqual(255, sunBigInt.FromHexString("0xfF"));
            AssertEqual(BigInteger.Parse("0DE", System.Globalization.NumberStyles.HexNumber), sunBigInt.FromHexString("0xDE"));
            AssertEqual(BigInteger.Parse("076AF390FFAE37BCDE01", System.Globalization.NumberStyles.HexNumber), sunBigInt.FromHexString("0x076AF390FFAE37BCDE01"));
            AssertEqual(BigInteger.Parse("01234567890abcdefABCDEF", System.Globalization.NumberStyles.HexNumber), sunBigInt.FromHexString("0x01234567890abcdefABCDEF"));

            AssertException<ArgumentException>(() => sunBigInt.FromHexString(""));
            AssertException<ArgumentException>(() => sunBigInt.FromHexString("0"));
            AssertException<ArgumentException>(() => sunBigInt.FromHexString("b"));
            AssertException<ArgumentException>(() => sunBigInt.FromHexString("-1"));
            AssertException<ArgumentException>(() => sunBigInt.FromHexString("121333g"));
            AssertException<ArgumentException>(() => sunBigInt.FromHexString("af0h1"));
            AssertException<ArgumentException>(() => sunBigInt.FromHexString(".12222"));

            testHandler.TestGroupEnd();
        }

        public void RunAdditionTestcases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Addition (+)");

            AssertEqual(new BigInteger(3) + new BigInteger(4), sunBigInt.FromLong(3) + sunBigInt.FromLong(4));
            AssertEqual(new BigInteger(0) + new BigInteger(1), sunBigInt.FromLong(0) + sunBigInt.FromLong(1));
            AssertEqual(new BigInteger(55555) + new BigInteger(0), sunBigInt.FromLong(55555) + sunBigInt.FromLong(0));
            AssertEqual(new BigInteger(98) + new BigInteger(98), sunBigInt.FromLong(98) + sunBigInt.FromLong(98));
            AssertEqual(new BigInteger(9711591238) + new BigInteger(91238),
                        sunBigInt.FromLong(9711591238) + sunBigInt.FromLong(91238));
            AssertEqual(new BigInteger(9999999999911111238) + new BigInteger(9999999999911123877),
                        sunBigInt.FromLong(9999999999911111238) + sunBigInt.FromLong(9999999999911123877));

            AssertEqual(BigInteger.Parse("345632221") + BigInteger.Parse("345631"),
                        sunBigInt.FromString("345632221") + sunBigInt.FromString("345631"));
            AssertEqual(BigInteger.Parse("45898736726236746427637632764874799124") + BigInteger.Parse("58362119983634542273"),
                        sunBigInt.FromString("45898736726236746427637632764874799124") + sunBigInt.FromString("58362119983634542273"));
            AssertEqual(BigInteger.Parse("64874799124") + BigInteger.Parse("83634542273"),
                        sunBigInt.FromString("64874799124") + sunBigInt.FromString("83634542273"));
            AssertEqual(BigInteger.Parse("104") + BigInteger.Parse("3434784378433434333223311111111119999999"),
                        sunBigInt.FromString("104") + sunBigInt.FromString("3434784378433434333223311111111119999999"));
            AssertEqual(BigInteger.Parse("104") + BigInteger.Parse("99999"),
                        sunBigInt.FromString("104") + sunBigInt.FromString("99999"));
            AssertEqual(BigInteger.Parse("37827382729877328327873287289723") + BigInteger.Parse("100000009"),
                        sunBigInt.FromString("37827382729877328327873287289723") + sunBigInt.FromString("100000009"));
            AssertEqual(BigInteger.Parse("0") + BigInteger.Parse("1000000009"),
                        sunBigInt.FromString("0") + sunBigInt.FromString("1000000009"));
            AssertEqual(BigInteger.Parse("8262110000025245610") + BigInteger.Parse("0"),
                        sunBigInt.FromString("8262110000025245610") + sunBigInt.FromString("0"));

            testHandler.TestGroupEnd();
        }

        public void RunSubtractionTestcases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Subtraction (-)");
            AssertEqual(new BigInteger(0) - new BigInteger(0), sunBigInt.FromLong(0) - sunBigInt.FromLong(0));
            AssertEqual(new BigInteger(1) - new BigInteger(0), sunBigInt.FromLong(1) - sunBigInt.FromLong(0));
            AssertEqual(BigInteger.Parse("4676376743829000019292") - BigInteger.Parse("4676376743829000019292"),
                        sunBigInt.FromString("4676376743829000019292") - sunBigInt.FromString("4676376743829000019292"));

            AssertEqual(new BigInteger(123) - new BigInteger(89), sunBigInt.FromLong(123) - sunBigInt.FromLong(89));
            AssertEqual(new BigInteger(123) - new BigInteger(99), sunBigInt.FromLong(123) - sunBigInt.FromLong(99));
            AssertEqual(new BigInteger(178) - new BigInteger(29), sunBigInt.FromLong(178) - sunBigInt.FromLong(29));
            AssertEqual(new BigInteger(583) - new BigInteger(299), sunBigInt.FromLong(583) - sunBigInt.FromLong(299));
            AssertEqual(new BigInteger(32232332) - new BigInteger(2500), sunBigInt.FromLong(32232332) - sunBigInt.FromLong(2500));
            AssertEqual(BigInteger.Parse("47865182653298722613") - BigInteger.Parse("3783786811999"),
                        sunBigInt.FromString("47865182653298722613") - sunBigInt.FromString("3783786811999"));

            AssertException<OverflowException>(() => { T a = sunBigInt.FromLong(0) - sunBigInt.FromLong(1); });
            AssertException<OverflowException>(() => { T a = sunBigInt.FromLong(99) - sunBigInt.FromLong(100); });
            AssertException<OverflowException>(() => { T a = sunBigInt.FromLong(45) - sunBigInt.FromLong(46); });

            testHandler.TestGroupEnd();
        }

        public void RunGreaterTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Greater (>)");

            AssertBool(false, COMPARISON_OPERATOR.GREATER, sunBigInt.FromLong(0), sunBigInt.FromLong(0));
            AssertBool(false, COMPARISON_OPERATOR.GREATER, sunBigInt.FromLong(1239), sunBigInt.FromLong(12341));
            AssertBool(false, COMPARISON_OPERATOR.GREATER, sunBigInt.FromLong(34433443), sunBigInt.FromLong(34433443));
            AssertBool(true, COMPARISON_OPERATOR.GREATER, sunBigInt.FromLong(34433444), sunBigInt.FromLong(34433443));
            AssertBool(true, COMPARISON_OPERATOR.GREATER, sunBigInt.FromLong(589), sunBigInt.FromLong(299));
            AssertBool(true, COMPARISON_OPERATOR.GREATER, sunBigInt.FromLong(289), sunBigInt.FromLong(288));
            AssertBool(true, COMPARISON_OPERATOR.GREATER, sunBigInt.FromLong(46467346764376), sunBigInt.FromLong(46467346758759));
            AssertBool(false, COMPARISON_OPERATOR.GREATER, sunBigInt.FromString("999000"), sunBigInt.FromString("9919999"));
            AssertBool(false, COMPARISON_OPERATOR.GREATER, sunBigInt.FromString("999828218"), sunBigInt.FromString("999828218"));
            AssertBool(true, COMPARISON_OPERATOR.GREATER, sunBigInt.FromString("3673665476377"), sunBigInt.FromString("3673665476376"));

            testHandler.TestGroupEnd();
        }

        public void RunGreaterThanTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("GreaterThan (>=)");

            AssertBool(true, COMPARISON_OPERATOR.GREATERTHAN, sunBigInt.FromLong(0), sunBigInt.FromLong(0));
            AssertBool(false, COMPARISON_OPERATOR.GREATERTHAN, sunBigInt.FromLong(1239), sunBigInt.FromLong(12341));
            AssertBool(true, COMPARISON_OPERATOR.GREATERTHAN, sunBigInt.FromLong(34433443), sunBigInt.FromLong(34433443));
            AssertBool(true, COMPARISON_OPERATOR.GREATERTHAN, sunBigInt.FromLong(34433444), sunBigInt.FromLong(34433443));
            AssertBool(true, COMPARISON_OPERATOR.GREATERTHAN, sunBigInt.FromLong(46467346764376), sunBigInt.FromLong(46467346758759));
            AssertBool(false, COMPARISON_OPERATOR.GREATERTHAN, sunBigInt.FromString("999000"), sunBigInt.FromString("9919999"));
            AssertBool(true, COMPARISON_OPERATOR.GREATERTHAN, sunBigInt.FromString("999828218"), sunBigInt.FromString("999828218"));
            AssertBool(true, COMPARISON_OPERATOR.GREATERTHAN, sunBigInt.FromString("3673665476377"), sunBigInt.FromString("3673665476376"));
            AssertBool(true, COMPARISON_OPERATOR.GREATERTHAN, sunBigInt.FromString("900002"), sunBigInt.FromString("900000"));

            testHandler.TestGroupEnd();
        }

        public void RunLessTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Smaller (<)");

            AssertBool(false, COMPARISON_OPERATOR.SMALLER, sunBigInt.FromLong(0), sunBigInt.FromLong(0));
            AssertBool(true, COMPARISON_OPERATOR.SMALLER, sunBigInt.FromLong(0), sunBigInt.FromLong(1));
            AssertBool(false, COMPARISON_OPERATOR.SMALLER, sunBigInt.FromLong(4444), sunBigInt.FromLong(4443));
            AssertBool(false, COMPARISON_OPERATOR.SMALLER, sunBigInt.FromLong(4444), sunBigInt.FromLong(4444));
            AssertBool(true, COMPARISON_OPERATOR.SMALLER, sunBigInt.FromLong(4444), sunBigInt.FromLong(44445));
            AssertBool(true, COMPARISON_OPERATOR.SMALLER, sunBigInt.FromLong(1234), sunBigInt.FromLong(12341));

            AssertBool(false, COMPARISON_OPERATOR.SMALLER, sunBigInt.FromString("4"), sunBigInt.FromString("4"));
            AssertBool(true, COMPARISON_OPERATOR.SMALLER, sunBigInt.FromString("4"), sunBigInt.FromString("5"));
            AssertBool(true, COMPARISON_OPERATOR.SMALLER, sunBigInt.FromString("4790001"), sunBigInt.FromString("4791000"));
            AssertBool(true, COMPARISON_OPERATOR.SMALLER, sunBigInt.FromString("900000"), sunBigInt.FromString("900002"));

            testHandler.TestGroupEnd();
        }

        public void RunLessThanTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("SmallerThan (<=)");

            AssertBool(true, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromLong(0), sunBigInt.FromLong(0));
            AssertBool(true, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromLong(0), sunBigInt.FromLong(1));
            AssertBool(false, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromLong(4444), sunBigInt.FromLong(4443));
            AssertBool(true, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromLong(4444), sunBigInt.FromLong(4444));
            AssertBool(true, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromLong(4444), sunBigInt.FromLong(44445));
            AssertBool(true, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromLong(1234), sunBigInt.FromLong(12341));

            AssertBool(true, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromString("4"), sunBigInt.FromString("4"));
            AssertBool(true, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromString("4"), sunBigInt.FromString("5"));
            AssertBool(true, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromString("4790001"), sunBigInt.FromString("4791000"));
            AssertBool(false, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromString("900001"), sunBigInt.FromString("900000"));
            AssertBool(false, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromString("632763763276736721191991"), sunBigInt.FromString("632763763276736721191989"));
            AssertBool(true, COMPARISON_OPERATOR.SMALLERTHAN, sunBigInt.FromString("321199127736311362398"), sunBigInt.FromString("321199127736312362398"));

            testHandler.TestGroupEnd();
        }

        public void RunEqualityThanTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Equality (==)");

            AssertBool(true, COMPARISON_OPERATOR.EQUALITY, sunBigInt.FromLong(0), sunBigInt.FromLong(0));
            AssertBool(true, COMPARISON_OPERATOR.EQUALITY, sunBigInt.FromLong(4343434391), sunBigInt.FromLong(4343434391));
            AssertBool(false, COMPARISON_OPERATOR.EQUALITY, sunBigInt.FromLong(4343434392), sunBigInt.FromLong(4343434391));
            AssertBool(false, COMPARISON_OPERATOR.EQUALITY, sunBigInt.FromLong(4443434391), sunBigInt.FromLong(4343434391));
            AssertBool(false, COMPARISON_OPERATOR.EQUALITY, sunBigInt.FromLong(200), sunBigInt.FromLong(202));

            AssertBool(true, COMPARISON_OPERATOR.EQUALITY, sunBigInt.FromString("0"), sunBigInt.FromString("0"));
            AssertBool(true, COMPARISON_OPERATOR.EQUALITY, sunBigInt.FromString("1367"), sunBigInt.FromString("1367"));
            AssertBool(false, COMPARISON_OPERATOR.EQUALITY, sunBigInt.FromString("1367"), sunBigInt.FromString("1376"));
            AssertBool(false, COMPARISON_OPERATOR.EQUALITY, sunBigInt.FromString("3443456577445454"), sunBigInt.FromString("3443456577445554"));

            testHandler.TestGroupEnd();
        }

        public void RunInequalityThanTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Inequality (==)");

            AssertBool(false, COMPARISON_OPERATOR.INEQUALITY, sunBigInt.FromLong(0), sunBigInt.FromLong(0));
            AssertBool(true, COMPARISON_OPERATOR.INEQUALITY, sunBigInt.FromLong(1), sunBigInt.FromLong(0));
            AssertBool(true, COMPARISON_OPERATOR.INEQUALITY, sunBigInt.FromLong(0), sunBigInt.FromLong(1));
            AssertBool(true, COMPARISON_OPERATOR.INEQUALITY, sunBigInt.FromLong(43344), sunBigInt.FromLong(53344));

            AssertBool(false, COMPARISON_OPERATOR.INEQUALITY, sunBigInt.FromString("0"), sunBigInt.FromString("0"));
            AssertBool(true, COMPARISON_OPERATOR.INEQUALITY, sunBigInt.FromString("0"), sunBigInt.FromString("2222"));
            AssertBool(true, COMPARISON_OPERATOR.INEQUALITY, sunBigInt.FromString("45"), sunBigInt.FromString("1"));
            AssertBool(true, COMPARISON_OPERATOR.INEQUALITY, sunBigInt.FromString("646565665332"), sunBigInt.FromString("646565665331"));

            testHandler.TestGroupEnd();
        }

        public void RunIsEvenTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Even");

            AssertBool(true, sunBigInt.FromLong(0).IsEven());
            AssertBool(false, sunBigInt.FromLong(1).IsEven());
            AssertBool(true, sunBigInt.FromString("3783782377376").IsEven());
            AssertBool(true, sunBigInt.FromString("1111111111111110").IsEven());
            AssertBool(false, sunBigInt.FromString("73867387879").IsEven());

            testHandler.TestGroupEnd();
        }

        public void RunIsOddTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Odd");

            AssertBool(false, sunBigInt.FromLong(0).IsOdd());
            AssertBool(true, sunBigInt.FromLong(1).IsOdd());
            AssertBool(false, sunBigInt.FromString("3783782377376").IsOdd());
            AssertBool(false, sunBigInt.FromString("1111111111111110").IsOdd());
            AssertBool(true, sunBigInt.FromString("73867387879").IsOdd());

            testHandler.TestGroupEnd();
        }

        public void RunMultplicationTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Multipication (*)");

            AssertEqual(new BigInteger(0) * new BigInteger(0), sunBigInt.FromLong(0).Mul(sunBigInt.FromLong(0)));
            AssertEqual(new BigInteger(0) * new BigInteger(19), sunBigInt.FromLong(0).Mul(sunBigInt.FromLong(19)));
            AssertEqual(new BigInteger(5643) * new BigInteger(0), sunBigInt.FromLong(5643).Mul(sunBigInt.FromLong(0)));
            AssertEqual(new BigInteger(1) * new BigInteger(378), sunBigInt.FromLong(1).Mul(sunBigInt.FromLong(378)));
            AssertEqual(new BigInteger(235409990009) * new BigInteger(1), sunBigInt.FromLong(235409990009).Mul(sunBigInt.FromLong(1)));
            AssertEqual(new BigInteger(109) * new BigInteger(45), sunBigInt.FromLong(109).Mul(sunBigInt.FromLong(45)));
            AssertEqual(new BigInteger(45) * new BigInteger(109), sunBigInt.FromLong(45).Mul(sunBigInt.FromLong(109)));
            AssertEqual(new BigInteger(99108) * new BigInteger(3448709199), sunBigInt.FromLong(99108).Mul(sunBigInt.FromLong(3448709199)));
            AssertEqual(new BigInteger(9) * new BigInteger(134), sunBigInt.FromLong(9).Mul(sunBigInt.FromLong(134)));
            AssertEqual(new BigInteger(134) * new BigInteger(9), sunBigInt.FromLong(134).Mul(sunBigInt.FromLong(9)));
            AssertEqual(new BigInteger(99) * new BigInteger(8765), sunBigInt.FromLong(99).Mul(sunBigInt.FromLong(8765)));
            AssertEqual(new BigInteger(8765) * new BigInteger(99), sunBigInt.FromLong(8765).Mul(sunBigInt.FromLong(99)));

            testHandler.TestGroupEnd();
        }

        public void RunDivisionTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Division (/)");

            AssertEqual(new BigInteger(100) / new BigInteger(5), sunBigInt.FromLong(100) / sunBigInt.FromLong(5));
            AssertEqual(BigInteger.Parse("244348463435788800") / BigInteger.Parse("7382549900"), sunBigInt.FromString("244348463435788800") / sunBigInt.FromString("7382549900"));
            AssertEqual(BigInteger.Parse("244348463435788800") / BigInteger.Parse("33098112"), sunBigInt.FromString("244348463435788800") / sunBigInt.FromString("33098112"));

            TestQuotientRemainder(new BigInteger(513), new BigInteger(13), sunBigInt.FromLong(513), sunBigInt.FromLong(13));
            TestQuotientRemainder(new BigInteger(3434), new BigInteger(3435), sunBigInt.FromLong(3434), sunBigInt.FromLong(3435));
            TestQuotientRemainder(BigInteger.Parse("154"), BigInteger.Parse("86290"), sunBigInt.FromString("154"), sunBigInt.FromString("86290"));

            TestQuotientRemainder(BigInteger.Parse("244348463435788800"), BigInteger.Parse("7382549901"),
                                  sunBigInt.FromString("244348463435788800"), sunBigInt.FromString("7382549901"));
            TestQuotientRemainder(BigInteger.Parse("244348463435788801"), BigInteger.Parse("7382549900"),
                                  sunBigInt.FromString("244348463435788801"), sunBigInt.FromString("7382549900"));
            TestQuotientRemainder(BigInteger.Parse("244348463435788799"), BigInteger.Parse("7382549900"),
                                  sunBigInt.FromString("244348463435788799"), sunBigInt.FromString("7382549900"));

            AssertException<DivideByZeroException>(() => { T a = sunBigInt.FromLong(4334335) / sunBigInt.FromLong(0); });
            testHandler.TestGroupEnd();
        }

        public void RunPowTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("Power (^^)");

            AssertEqual(BigInteger.Pow(BigInteger.Parse("123456789"), 0), sunBigInt.Pow(sunBigInt.FromString("123456789"), 0));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("123456789"), 1), sunBigInt.Pow(sunBigInt.FromString("123456789"), 1));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("2"), 5), sunBigInt.Pow(sunBigInt.FromString("2"), 5));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("123456789"), 5), sunBigInt.Pow(sunBigInt.FromString("123456789"), 5));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("2"), 10), sunBigInt.Pow(sunBigInt.FromString("2"), 10));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("2"), 11), sunBigInt.Pow(sunBigInt.FromString("2"), 11));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("23"), 11), sunBigInt.Pow(sunBigInt.FromString("23"), 11));
            AssertEqual(BigInteger.Pow(BigInteger.Parse("32327"), 134), sunBigInt.Pow(sunBigInt.FromString("32327"), 134));

            testHandler.TestGroupEnd();
        }

        public void RunModPowTestCases<T>(T sunBigInt) where T : SunBigUIntAbstract<T>
        {
            testHandler.TestGroupStart("ModPow");

            AssertEqual(BigInteger.ModPow(4, 13, 497), sunBigInt.ModPow(4, 13, 497));
            AssertEqual(BigInteger.ModPow(23, 12, 10098), sunBigInt.ModPow(23, 12, 10098));
            AssertEqual(BigInteger.ModPow(2, 45, 123), sunBigInt.ModPow(2, 45, 123));
            AssertEqual(BigInteger.ModPow(103, 56, 7), sunBigInt.ModPow(103, 56, 7));
            AssertEqual(BigInteger.ModPow(200, 5, 1), sunBigInt.ModPow(200, 5, 1));

            testHandler.TestGroupEnd();
        }

        private void TestQuotientRemainder<T>(BigInteger bigInt1, BigInteger bigInt2, T sunInt1, T sunInt2) where T : SunBigUIntAbstract<T>
        {
            BigInteger bigIntRemainder, bigIntQuotient;
            SunBigUIntAbstract<T> sunRemainder, sunQuotient;

            bigIntQuotient = BigInteger.DivRem(bigInt1, bigInt2, out bigIntRemainder);
            sunQuotient = sunInt1.DivRem(sunInt1, sunInt2, out sunRemainder);

            AssertEqual(bigIntQuotient, (T)sunQuotient);
            AssertEqual(bigIntRemainder, (T)sunRemainder);
        }

        private void AssertException<T>(Action a) where T : Exception
        {
            try
            {
                a.Invoke();
                this.testHandler.TestStepFail("No exception occured.");
            }
            catch (T)
            {
                this.testHandler.TestStepSuccess();
            }
            catch (Exception)
            {
                this.testHandler.TestStepFail("Wrong exception occured.");
            }

        }

        private void AssertEqual<T>(BigInteger expectedInt, T actualInt) where T : SunBigUIntAbstract<T>
        {
            string expIntStr = expectedInt.ToString();
            string actIntStr = actualInt.ToString();

            if (expIntStr.Equals(actIntStr))
            {
                this.testHandler.TestStepSuccess();
            }
            else
            {
                this.testHandler.TestStepFail(expectedInt, actualInt);
                NumOfFailedTestcases++;
            }
        }

        private void AssertBool<T>(bool expectedResult, COMPARISON_OPERATOR c, T a, T b) where T : SunBigUIntAbstract<T>
        {
            if (c == COMPARISON_OPERATOR.GREATER)
            {
                AssertBool(expectedResult, a > b, a, b);
            }
            else if (c == COMPARISON_OPERATOR.GREATERTHAN)
            {
                AssertBool(expectedResult, a >= b, a, b);
            }
            else if (c == COMPARISON_OPERATOR.SMALLER)
            {
                AssertBool(expectedResult, a < b, a, b);
            }
            else if (c == COMPARISON_OPERATOR.SMALLERTHAN)
            {
                AssertBool(expectedResult, a <= b, a, b);
            }
            else if (c == COMPARISON_OPERATOR.EQUALITY)
            {
                AssertBool(expectedResult, a == b, a, b);
            }
            else if (c == COMPARISON_OPERATOR.INEQUALITY)
            {
                AssertBool(expectedResult, a != b, a, b);
            }
        }

        private void AssertBool<T>(bool expectedResult, bool actualResult, T a, T b) where T : SunBigUIntAbstract<T>
        {
            if (expectedResult == actualResult)
            {
                this.testHandler.TestStepSuccess();
            }
            else
            {
                this.testHandler.TestStepFail("Expected: " + expectedResult.ToString() + ", Actual " + actualResult.ToString() +
                                              "(a = " + a.ToString() + ", b = " + b.ToString() + ")");
                NumOfFailedTestcases++;
            }
        }

        private void AssertBool(bool expectedResult, bool actualResult)
        {
            if (expectedResult == actualResult)
            {
                this.testHandler.TestStepSuccess();
            }
            else
            {
                this.testHandler.TestStepFail("Expected: " + expectedResult.ToString() + ", Actual " + actualResult.ToString());
                NumOfFailedTestcases++;
            }
        }

    }
}
